"""Module to run minimal flask app to be used for tests.
"""

from flask import Flask
APP = Flask(__name__)

@APP.route('/')
def hello_world():
    return 'Hello, World!'

@APP.route('/foo')
def foo():
    return 'bar'



"""Useful decorators.
"""

import os
import json
import uuid
import logging
import datetime
import time
import threading

from contextlib import ContextDecorator

import wrapt

from flask import request, g


def _start_watch(name, w_type, args, kwargs, show_args=False):
    """Helper function to start watching command

    :param name:    String name of command to watch.

    :param w_type:  String type (e.g., 'function', 'flask_request').

    :param args:    Arguments to command.

    :param kwargs:  Keyword args.

    :param show_args=False:  Whether to show args/kwargs in log.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Dictionary of meta data for watched command.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Create dictionary of meta data and log it so we can
              watch a command. Meant to be called by watched decorator.

    """

    w_data = {'w_uuid': str(uuid.uuid4()),  # UUID to help tracking.
              'w_name': name, 'watched': 'start', 'w_type': w_type,
              'start': time.time()}
    if show_args:
        w_data.update(w_args=args, w_kwargs=str(kwargs))
    logging.info('watched_cmd: %s', name, extra=w_data)
    return w_data


def _end_watch(w_data, str_result):
    """Log an error in a watched command.

    :param w_data:    As produced by _start_watch.

    :param str_result:  Result of command as a string.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Updated `w_data`.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Log about a watched command ending.
              Helper meant to be called only be `watched` decorator.

    """

    cmd_time = time.time() - w_data['start']
    w_data.update(watched='end', status='ok', w_run_time=cmd_time,
                  w_result=str_result)
    logging.info('watched_cmd_end: ok:%s', w_data['w_name'], extra=w_data)
    return w_data


def _error_watch(w_data, my_problem):
    """Log an error in a watched command.

    :param w_data:    As produced by _start_watch.

    :param my_problem:    Problem description.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Updated `w_data`.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Log a warning about an error in a watched command.
              Helper meant to be called only be `watched` decorator.

    """
    cmd_time = time.time() - w_data['start']
    w_data.update(w_error=str(my_problem), status='error',
                  w_run_time=cmd_time, watched='end')
    logging.warning('watched_end_cmd: error:%s', w_data['w_name'],
                    extra=w_data)
    return w_data


@wrapt.decorator
def watched(wrapped, instance, args, kwargs, show_args=True):
    """Decorator to make a command "watched" where we track timing.

    :param wrapped:    Function to wrap.

    :param instance:   Generally ignored (can be instance of class).

    :param args, kwargs:  Arguments passed to wrapped function. If
                          `show_args` is True, then we log these.

    :param show_args=True:  Whether to show args in logs.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Same as calling wrapped function.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Decorator to make a "watched" function where we put info
              about start, end, error, args, into logs. This will create
              log messages starting with 'watched_*' and providing an
              `extra` dictionary to logging with useful meta data. You can
              then search your logs for things like 'watched_end_cmd' to
              find info on how the command ran.

              This is especially useful with something like logz.io.
    """
    logging.debug('watched_meta: Ignoring instance %s', instance)
    name = wrapped.__name__
    w_data = _start_watch(name, 'function', args, kwargs, show_args)
    try:
        str_result = '(unknown)'
        result = wrapped(*args, **kwargs)
        try:
            str_result = str(result)
            if len(str_result) > 200:
                str_result = str_result[:195]+'...'
        except Exception as unexpected:  # pylint: disable=broad-except
            logging.error('Ignoring unexpected str conversion exception: %s',
                          unexpected)
        _end_watch(w_data, str_result)
        return result
    except Exception as my_problem:  # pylint: disable=broad-except
        _error_watch(w_data, my_problem)
        raise


def _start_watching_request(req=None, name=None):
    """Start watching a flask request.

Uses the `_start_watch`, `_end_watch` helpers to watch a flask
command. Use `setup_flask_watch` to turn on for your app. See
docs for `setup_flask_watch` for details.
    """
    req = req or request
    name = name or request.endpoint
    w_data = _start_watch(name, 'flask_request', None, None, False)
    g.w_data = w_data  # pylint: disable=assigning-non-slot


def _end_watching_request(req=None):
    """End watching a flask request.

Uses the `_start_watch`, `_end_watch` helpers to watch a flask
command. Use `setup_flask_watch` to turn on for your app. See
docs for `setup_flask_watch` for details.
    """
    req = req or request
    w_data = getattr(g, 'w_data', None)
    if w_data is None:
        logging.error('No w_data member for request: %s', req)
        return
    str_result = ''
    _end_watch(w_data, str_result[:195])


@watched
def setup_flask_watch(app):
    """Take a flask app as input and setup logs to watch requests.

This uses tools similar to the `watched` decorator to watch and log
start/end of a flask command using before_request and teardown_request
hooks.

Basically just call this on your app and then search your logs for
something like `watched_cmd*` to see info about your flask commands.
    """
    app.before_request(_start_watching_request)
    app.teardown_request(_end_watching_request)


class LockFile(ContextDecorator):
    """Context decorator to create/check lock files.

This class can serve as a decorator or a context manager to
create/check a lock file.

The following illustrates example usage:

>>> import time, threading, os, logging
>>> from ox_ui.core import decorators
>>> lock_path = '/tmp/test.lock'     # example lock file path
>>> @decorators.LockFile(lock_path)  # decorate function to use lock file
... def foo(t, m):
...     assert os.path.exists(lock_path)
...     logging.warning('%s Sleep %s thread %s', m, t, threading.get_ident())
...     time.sleep(t)
...
>>> my_thread = threading.Thread(target=foo, args=[5, 'thread'])
>>> problems = []  # Thread is started in background.
>>> try:           # Verify a FileExistsError raised if conflict occurs.
...     my_thread.start()  # start a background thread with the lock
...     while not os.path.exists(lock_path):
...         time.sleep(0.5)
...     foo(4, 'main')     # now spawn in main thread to poke lock
... except Exception as problem:
...     problems.append(problem)
...
>>> assert len(problems) > 0, problems  # Verify we caught an exception.

    """

    def __init__(self, lockpath, comment='', encoding='utf8'):
        self.lockpath = lockpath
        self.encoding = encoding
        self.comment = comment
        self.created = None

    def remove_lock(self):
        """Remove lock if it exists.
        """
        if os.path.exists(self.lockpath):
            os.remove(self.lockpath)

    def __enter__(self):
        if os.path.exists(self.lockpath):
            info = 'unknown'
            try:
                with open(self.lockpath, encoding=self.encoding) as fdesc:
                    info = json.load(fdesc)
                logging.warning('Found lock file %s with data: %s',
                                self.lockpath, info)
            except Exception:  # pylint: disable=broad-except
                logging.exception('Unable to get info about lock file %s',
                                  self.lockpath)
            raise FileExistsError(self.lockpath)
        with open(self.lockpath, 'w', encoding=self.encoding) as fdesc:
            self.created = datetime.datetime.now()
            info = {'pid': os.getpid(), 'comment': self.comment,
                    'thread_id': threading.get_ident(),
                    'created_dt': str(self.created),
                    'created_ts': self.created.timestamp()}
            json.dump(info, fdesc, indent=2)
        return self

    def __exit__(self, exc_type, exc, exc_tb):
        if exc_type:
            logging.debug('Ignoring exception info %s, %s, %s',
                          exc_type, exc, exc_tb)
        self.remove_lock()
        return False



"""Test some basic things.
"""

import datetime
import unittest
import click

from ox_ui.core import c2g, decorators


@click.command()
@click.option('--name', '-n', default='you',
              help='Name of person to hello.')
@click.option('--count', '-c', help='How many times to repeat.',
              default=1)
@click.option('--when', '-w', type=click.DateTime())
@click.option('--dry-run/--no-dry-run', default=False, help=(
    'If dry-run then do nothing and return ["dry-run"]'))
def hello_world(name, count, when, dry_run):
    "Say hello to person multiple times."

    if dry_run:
        return ["dry-run"]
    if when is None:
        result = []
    else:
        result = ['At ' + str(when.date()) + ': ']
    result.extend(['Hello ' + name]*count)
    return result


class TestHelloWorld(unittest.TestCase):
    """Test simple hello_world method.
    """

    @decorators.watched  # decorator is optional
    def test_when(self):
        "Test with a `when` argument."
        hw_cmd = c2g.ClickToGeneric(hello_world)
        result = hw_cmd.handle_request({
            'when': datetime.datetime(2021, 2, 3, 4, 5, 6)})
        self.assertEqual(result, ['At 2021-02-03: ', 'Hello you'])

    def test_default(self):
        "Test with default arguments."
        hw_cmd = c2g.ClickToGeneric(hello_world)
        result = hw_cmd.handle_request()
        self.assertEqual(result, ['Hello you'])

    def test_notyou(self):
        "Test with name = 'notyou'."
        hw_cmd = c2g.ClickToGeneric(hello_world)
        result = hw_cmd.handle_request({'name': 'notyou'})
        self.assertEqual(result, ['Hello notyou'])

    def test_you_2(self):
        "Test with a count = 2 argument."
        hw_cmd = c2g.ClickToGeneric(hello_world)
        result = hw_cmd.handle_request({'count': 2})
        self.assertEqual(len(result), 2)

    def test_dry_run(self):
        "Test dru run argument."
        hw_cmd = c2g.ClickToGeneric(hello_world)
        result = hw_cmd.handle_request({'dry_run': True})
        self.assertEqual(result, ['dry-run'])


if __name__ == '__main__':
    print('Runnig tests')
    unittest.main()



"""Tools to convert click commands to generic command
"""

import datetime
import re
import logging

from click import types


class GenericField:
    """Class representing generic field similar to Flask WTForms
    """

    def __init__(self, name, validators=(), description=None,
                 type=str,  # pylint: disable=redefined-builtin
                 default=None):
        self.name = name
        self.validators = validators
        self.description = description
        self.data = default
        self.type = type


class ClickToGeneric:
    """Class to convert click command to generic command.
    """

    def __init__(self, click_cmd, skip_opt_re=None, tweaks: list = None):
        """Initializer.

        :param click_cmd:   A click command to turn into a form.

        :param skip_opt_re=None:  Regexp for options to skip in form.

        :param tweaks:  List of 'tweaks' to add (e.g., FileResponseTweak).

        """
        self.click_cmd = click_cmd
        self.skip_opt_re = skip_opt_re if not skip_opt_re else re.compile(
            skip_opt_re)
        self.tweaks = tweaks if tweaks else []
        self.gobbled_opts = {}

    def form_cls(self):
        """Create a class to represent the generic form for the command.
        """

        class GenericForm:
            """Generic form to run click command
            """

        for opt in self.click_cmd.params:
            if self.skip_opt_re and self.skip_opt_re.search(opt.name):
                logging.info('Option %s since matchs skip_opt_re', opt.name)
            elif self.gobble(opt.name):
                logging.info('Option %s gobbled', opt.name)
            else:
                field = self.click_opt_to_field(opt)
                setattr(GenericForm, opt.name, field)

        return GenericForm

    def gobble(self, name):
        """Remove options from standard processing.

This is needed so we can handle some special types of options differently.
        """
        for tweak in self.tweaks:
            reason = tweak.gobble(self, name)
            if reason:
                self.gobbled_opts[name] = reason
                return reason
        return None

    def pad_kwargs(self, kwargs):
        """Handle keyword args that need padding.
        """
        for tweak in self.tweaks:
            tweak.pad_kwargs(self, kwargs)

    def form(self, opts):
        """Create an instance of a generic form showing command arguments.
        """
        cls = self.form_cls()
        my_form = cls()
        for name, value in opts.items():
            if value is not None:
                field = getattr(my_form, name)
                field.data = value

        return my_form

    def make_int_field(self, opt):
        "Make an integer field."

        logging.debug('Making field for %s', self)
        return GenericField(
            opt.name, description=opt.help, default=opt.default,
            type=int)

    def make_bool_field(self, opt):
        "Make a bool field."

        logging.debug('Making field for %s', self)
        return GenericField(
            opt.name, description=opt.help, default=opt.default,
            type=bool)

    def make_str_field(self, opt):
        "Make a string field."

        logging.debug('Making field for %s', self)
        return GenericField(
            opt.name, description=opt.help, default=opt.default,
            type=int)

    def make_dt_field(self, opt):
        "Make a datetime field."

        logging.debug('Making field for %s', self)
        default = opt.default
        if callable(default):
            default = default()
        if isinstance(default, str):
            default = datetime.datetime.strptime(
                default, '%Y-%m-%d %H:%M:%S')

        return GenericField(
            opt.name, description=opt.help, default=default,
            type=datetime.datetime)

    def make_file_field(self, opt):
        "Make a file field."

        logging.debug('Making field for %s', self)
        return GenericField(opt.name, description=opt.help,
                            default=opt.default)

    def click_opt_to_field(self, opt):
        """Convert given click option to a GenericField.
        """
        if opt.type == types.INT:
            field = self.make_int_field(opt)
        elif opt.type == types.BOOL:
            field = self.make_bool_field(opt)
        elif opt.type == types.STRING:
            field = self.make_str_field(opt)
        elif isinstance(opt.type, types.DateTime) or (
                getattr(opt.type, 'name', '?') == 'datetime'):
            field = self.make_dt_field(opt)
        elif isinstance(opt.type, (types.File, types.Path)):
            field = self.make_file_field(opt)
        else:
            raise TypeError(f'Cannot represent click type {opt.type}')
        return field

    def process(self, form):
        """Take a generic form instance as input and process the command.

        :param form:    GenericForm instance with command options/args.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   Result of running the command.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   This handles figuring out the arguments, extracting
                   the callback from click, and executing it.
        """
        kwargs = {}
        for opt in self.click_cmd.params:
            if opt.name in self.gobbled_opts:
                continue
            value = getattr(form, opt.name).data
            if opt.multiple:
                cur = kwargs.get(opt.name, [])
                if not cur:
                    kwargs[opt.name] = cur
                cur.append(value)
            else:
                kwargs[opt.name] = value

        self.pad_kwargs(kwargs)
        wrapped_cmd = getattr(self.click_cmd.callback, '__wrapped__', None)
        if wrapped_cmd:
            raw_result = wrapped_cmd(**kwargs)
        else:
            raw_result = self.click_cmd.callback(**kwargs)

        return self.post_process(raw_result)

    def post_process(self, raw_result):
        """Take output of process method and do further postprocessing.

        Intended for sub-classes to override.
        """
        logging.debug('For %s, post-processing raw_result %s',
                      self, raw_result)
        result = raw_result
        for tweak in self.tweaks:
            result = tweak.post_process_result(self, result)
        return result

    def handle_request(self, opts=None):
        """Handle request to run command:

        :param opts=None:   Optional dictionary with string keys for names
                            of options to command and values to use.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  Result of running the command.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Run the command in a generic way.

        """
        opts = opts or {}
        form = self.form(opts)
        result = self.process(form)
        return result



"""Test using flask server.
"""

import logging
import os
import socket
import time
import pathlib
import unittest
import tempfile

import requests

import ox_ui
from ox_ui.test_tools import server_tools


class BasicC2FTest(unittest.TestCase):

    server_info = None

    @classmethod
    def setUpClass(cls):
        cmd = ['flask', 'run']
        cwd = pathlib.Path(ox_ui.__file__).parents[1].joinpath(
            'tests', 'flask')
        env = os.environ.copy()
        env['FLASK_APP'] = str(cwd.joinpath('c2f_app.py'))
        cls.server_info = server_tools.start_server(cmd=cmd, cwd=cwd, env=env)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        for attempt in range(10):
            result = sock.connect_ex(('127.0.0.1', cls.server_info.s_port))
            if result == 0:
                logging.info('Succesfully started server at port %s',
                             cls.server_info.s_port)
                break
            else:
                logging.warning('Sleeping 2**%i to wait for server', attempt)
                time.sleep(2**i)
        else:
            raise ValueError('Could not start flask server.')

    @classmethod
    def tearDownClass(cls):
        s_proc = cls.server_info.s_proc
        if s_proc:
            s_proc.kill()
            s_proc.wait(timeout=10)
            cls.server_info.s_proc = None

    def test_count_file_size(self):
        url = 'http://127.0.0.1:%s/count_file_size' % (self.server_info.s_port)
        with tempfile.TemporaryDirectory() as tmpdir:
            datafile = tmpdir + '/data.txt'
            data = 'example\n'
            with open(datafile, 'w') as fdesc:
                fdesc.write(data)
            with open(datafile, 'r') as fdesc:
                req = requests.post(url, files=[('datafile', fdesc)])
                self.assertEqual(req.status_code, 200)
                self.assertEqual(req.text, str(len(data)))

    def test_choice(self):
        "Test if the choice mapping works right."

        url = 'http://127.0.0.1:%s/favorite_color' % (
            self.server_info.s_port)
        post_req = requests.post(url, data={'color': 'green'})  # default
        self.assertEqual(post_req.status_code, 200)
        self.assertEqual(post_req.text, 'I like green as well.')
        post_req = requests.post(url, data={'color': 'red'})  # choose red
        self.assertEqual(post_req.status_code, 200)
        self.assertEqual(post_req.text, 'I like red as well.')

    def test_good_c2f(self):
        url = 'http://127.0.0.1:%s/hello' % (self.server_info.s_port)
        count, text = 3, 'bye'
        self.check_good_c2f(url, count, text)

        count, text = 2, 'why'
        url = 'http://127.0.0.1:%s/goodbye' % (self.server_info.s_port)
        self.check_good_c2f(
            url, count, text, when='2021-01-23', also='blah',
            more='\nblah\n\nat 2021-01-23 00:00:00')

    def check_good_c2f(self, url, count, text, more='', **extras):

        data = {'count': count, 'text': text}
        data.update(extras)
        get_req = requests.get(url)
        self.assertEqual(get_req.status_code, 200)
        post_req = requests.post(url, data=data)
        self.assertEqual(post_req.status_code, 200)
        self.assertEqual(post_req.text, '\n'.join([text]*count) + more)

    def test_bad_c2f(self):
        url = 'http://127.0.0.1:%s/hello' % (self.server_info.s_port)
        get_req = requests.get(url)
        self.assertEqual(get_req.status_code, 200)
        count, text = 'bad', 'bye'
        post_req = requests.post(url, data={'count': count, 'text': text})
        self.assertEqual(post_req.status_code, 200)



"""Module to run minimal flask app to be used for tests.

You can do something like the following to serve this app
if you want to test/debug manually:

  FLASK_APP=c2f_app.py python3 -m flask run

"""

import datetime

import click
from flask import Flask, url_for

from ox_ui.core import c2f, decorators


APP = Flask(__name__)
APP.config['WTF_CSRF_ENABLED'] = False
decorators.setup_flask_watch(APP)


@click.command()
@click.option('--count', default=1, type=int, help='how many times to say it')
@click.option('--text', default='hi', type=str, help='what to say')
def hello_cmd(count, text):
    'say hello'

    result = []
    for i in range(count):
        result.append(text)

    return '\n'.join(result)


@click.command()
@click.option('--count', default=2, type=int, help='how many times to say it')
@click.option('--when', type=click.DateTime(formats=[
    '%Y-%m-%d', '%m/%d/%Y']), default=datetime.date.today)
@click.option('--text', default='bye', type=str, help='what to say')
@click.option('--also', multiple=True, default=['them'], type=str,
              help='Also say...; can be provided multiple times.')
def goodbye_cmd(count, text, when, also):
    'say bye'

    result = []
    for i in range(count):
        result.append(text)
    if also:
        result.extend(also)
    result.append('\nat %s' % str(when))

    return '\n'.join(result)


@click.command()
@click.option('--datafile', type=click.File('rb'), help=(
    'Data file to read.'))
def count_file_size_cmd(datafile):
    "Count size of file."

    result = len(datafile.read())
    return str(result)


@click.command()
@click.option('--color', type=click.Choice(['red', 'green', 'blue']), help=(
    'Favorite color.'), default='green')
def favorite_color_cmd(color):
    "Tell me your favorite color."

    return f'I like {color} as well.'


@APP.route('/')
def home():
    url = url_for('hello')
    return 'Welcome to test; try the view at <A HREF="%s">%s</A>' % (
        url, url)


@APP.route('/hello', methods=('GET', 'POST'))
def hello():
    fcmd = c2f.ClickToWTF(hello_cmd)
    result = fcmd.handle_request()
    return result


@APP.route('/goodbye', methods=('GET', 'POST'))
def goodbye():
    return c2f.ClickToWTF(goodbye_cmd).handle_request()


@APP.route('/count_file_size', methods=('GET', 'POST'))
def count_file_size():
    return c2f.ClickToWTF(count_file_size_cmd).handle_request()


@APP.route('/favorite_color', methods=('GET', 'POST'))
def favorite_color():
    return c2f.ClickToWTF(favorite_color_cmd).handle_request()
